/*
 *    PM6025.cpp - A library for the DisplayTech PM6025 LCD 
 *    Copyright (c) 2016 Istvan Cserny
 * 
 *    Permission is hereby granted, free of charge, to any person
 *    obtaining a copy of this software and associated documentation
 *    files (the "Software"), to deal in the Software without
 *    restriction, including without limitation the rights to use,
 *    copy, modify, merge, publish, distribute, sublicense, and/or sell
 *    copies of the Software, and to permit persons to whom the
 *    Software is furnished to do so, subject to the following
 *    conditions:
 * 
 *    This permission notice shall be included in all copies or 
 *    substantial portions of the Software.
 * 
 *    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 *    OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 *    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 *    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 *    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 *    OTHER DEALINGS IN THE SOFTWARE.
 */

#include "PM6025.h"
#include <Wire.h>

/** 
 *   Constructor - nothing to do.. 
 */
PM6025::PM6025() {
	Wire.begin();
}

/**********************************************************
 *       P R I V A T E   F U N C T I O N S                *
 *********************************************************/
 
/**
 *   Get a half byte from the status[] array and returns it 
 *   in the lower half of a byte
 */
uint8_t PM6025::getNibble(uint8_t addr) {

	uint8_t idx = addr/2;
	uint8_t data = status[idx];
	if(addr&0x01) {
			return (data&0x0F);
	} else {
			return (data>>4);
	}
}

/**
 *   Fill a given half byte in the status[] array with the given data
 */
void PM6025::setNibble(uint8_t addr, uint8_t data) {
	uint8_t idx;
	idx = addr/2;
	data = data & 0x0F;
	if(addr&0x01) {	
		status[idx] = status[idx]&0xF0 | data;		
	} else {
		status[idx] = status[idx]&0x0F | data<<4;			
	}
}

/**
 *  Update the PM6025 display from data stored in the status[] array
 */
void PM6025::flush() {				
	Wire.beginTransmission(0x38);	
	Wire.write(0xC8);				//Set display mode
	Wire.write(0xF0);               //Set noblink mode
	Wire.write(0xE0);               //First block of data (20 bytes)
	Wire.write(0);
	Wire.write(status,20);
	Wire.endTransmission();	
	Wire.beginTransmission(0x38);		
	Wire.write(0xE1);              //Second block of data (20 bytes)
	Wire.write(0);
	Wire.write(&status[20],20);
	Wire.endTransmission();	
	Wire.beginTransmission(0x38);		
	Wire.write(0xE2);              //Third block of data (20 bytes)
	Wire.write(0);
	Wire.write(&status[40],20);	
	Wire.endTransmission();	
}	

/********************************************************
 *     P U B L I C   F U N C T I O N S                  *
 *******************************************************/

/**
 *   LCD initialization: test mode, then clear display 
 */
void PM6025::init(uint8_t data) {
	for(int i=0; i<60; i++) {
		status[i] = data;
	}
	flush();
}

/** 
 * Display a 7-segment digit in the given position
 * Only hexadecimal digits (0-F) and SPACE make sense here
 * Params:
 * row		row number (1 or 2)
 * pos		the position of the character on the display (1..13)
 * c		the digit to be displayed (0..16) 
 * Note: SPACE (c=16) and MINUS_SIGN (c=17) was also introduced for convenience
 */
void PM6025::setDigit(uint8_t row, uint8_t pos, uint8_t c) {	
	uint8_t value, addr;
	if(row < 1 || row > 2)  return; 		//Row 1 or 2 is allowed
	if(pos < 1 || pos > 13) return;			//Check for valid character position (1-13)
	if(c < 0 || c > 17) return;			    //Check for a valid digit value (0-17)
	value = numTable[c];
	addr = row1[pos-1] - ((row-1)<<1);		//Calculate array index	
	setNibble(addr++,value>>4);
	setNibble(addr,value&0x0F);
	flush();
}
		
/** 
 * Display a character on the 3rd row (14 character 16-segment display).
 * Only numbers (0-9), SPACE and uppercase letters (A-Z) make sense here.
 * Params:
 * pos		the position of the character on the display (1..14)
 * value	the code of the character to be displayed ('0'..'9', '@'..'Z', SPACE) 
 * 
 * Example:
 *   setChar(14,'A');		//Write letter "A" into the leftmost position of row 3
 */
void PM6025::setChar(uint8_t pos, char c) {
	uint16_t value;
	if(pos < 1 || pos > 14) return;			//Check for valid character position (1-14)
	if(c==' ') c = ':';						//Space equivalent for any non-printing character
	if(c < '0' || c > 'Z') return;			//Check for a valid character code ('0' - 'Z')
	int addr = 28 + (pos<<1);				//Calculate array index
    value = charTable[c-48];				//Get 16 segments data from table 
	status[addr++] = value>>8;				//Set abc segments in status[] array
	status[addr++] = value & 0xFF;			//Set fged segments in status[] array
	flush();								//Flush updated data to LCD
}

/**
 *  Sets sidebar symbol controlling bits in status[] array.
 *   
 *  Role of bits in parameter 'flags': 
 *  ---------------------------------------------------------------------------------
 *  | 15 | 14 | 13 | 12 | 11 | 10 |  9 |  8 |  7 |  6 |  5 |  4 |  3 |  2 |  1 |  0 | 
 *  ---------------------------------------------------------------------------------
 *  | S1 | S2 | S3 | S4 |  - |  - |  - |  - | V1 | V2 | V3 | V4 |  - |  - |  - | V5 | 
 *  ---------------------------------------------------------------------------------
 *
 *  Example:
 *    setSymbols(sBalance | sBattery | V1);  //Sets S3, S4 and V1 symbols.
 */
void PM6025::setSymbols(uint16_t flags) {
	status[58] = (status[58] & 0x0F) | (highByte(flags) & 0xF0);
	status[59] = lowByte(flags);
	flush();
}

/** 
 * Display decimal comma at the given position
 * Params:
 * row		row number (1 or 2) 
 * pos		character position (1..10, except of 7)
 * 
 * Example:
 *   setDP(1,5);		//Diplay decimal comma in row 1 after the 5th character (counting from right to left)
 */	
void PM6025::setDP(uint8_t row, uint8_t pos) {  
	uint8_t pv_data, pv_idx, pv_bits;
	if(row < 1 || row > 2)  return; 		//Row 1 or 2 is allowed
	if(pos < 1 || pos > 10 || pos == 7)		//Check for valid character position (1-10)	
		return;
	pv_idx = pvpv[pos-1];					//Find nibble positon
	pv_data = getNibble(pv_idx);			//get nibble form status[] array
    pv_bits = (row==1)?0x0C : 0x03;			//Bit position depends on row number
	setNibble(pv_idx,pv_data | pv_bits);	//set bits to display a comma
	flush();								//refresh screen
}

/** 
 * Remove (clear) decimal comma at the given position
 * Params:
 * row		row number (1 or 2) 
 * pos		character position (1..10, except of 7)
 * 
 * Example:
 *   setDP(1,5);		//Clear decimal comma in row 1 after the 5th character (counting from right to left)
 */	
void PM6025::clearDP(uint8_t row, uint8_t pos) {  
	uint8_t pv_data, pv_idx, pv_bits;
	if(row < 1 || row > 2)  return; 		//Row 1 or 2 is allowed
	if(pos < 1 || pos > 10 || pos == 7)		//Check for valid character position (1-10)	
		return;
	pv_idx = pvpv[pos-1];					//Find nibble positon
	pv_data = getNibble(pv_idx);			//get nibble form status[] array
    pv_bits = (row==1)?0x0C : 0x03;			//Bit position depends on row number
	setNibble(pv_idx,pv_data & ~pv_bits);	//clear bits to remove a comma
	flush();								//refresh screen
}

/**
 *   Display signed integers
 *   Parameters:
 *   	row		row number (1 or 2)
 * 		pos		the position of the character on the display (1..13)
 * 		data	number to be displayed (0..16) 
 */
void PM6025::write(uint8_t row, uint8_t pos, long data) {
	uint8_t sign=16;
	if(data<0) {
		sign=17;
		data = - data;
	}	
	do {
		setDigit(row, pos, (uint8_t)(data%10));
		data=data/10;
		pos++;
	} while(data>0);
	setDigit(row, pos, sign);    
}
